/******************************************************************************
*
* project name:    SNOW
* initial date:    26/06/2000
* author:          thomas.nussbaumer@peem.com
* description:     snow effect demonstration
*
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
* NOTE: depends on TIGCC Standard Library 2.x to compile
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*
* compile with: tigcc -O2 snow.c
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.00: 26/06/2000
*        initial version
*
*******************************************************************************/

#define SAVE_SCREEN      // restore screen after program end

#include <nostub.h>
#include <all.h>

/*---------------------------------------------------------------------------*/
/* some constants                                                            */
/*                                                                           */
/* NOTE: This demo is coded to run on both calculator types (TI89 & TI92)    */
/*       and uses therefore the screen extensions of the larger one (TI92).  */
/*       Due to that fact pixels on the TI89 are treated which aren't        */
/*       visible !!                                                          */
/*---------------------------------------------------------------------------*/
#define NR_PARTICLES  500      // number of max. snow particles to handle
#define RAND_RANGE    240*126  // range in which pixels are selected 
                               // (== TI92 screen extensions !!)
#define FREE_ENTRY    0xffff   // indicates a not used particle entry


/*---------------------------------------------------------------------------*/
/* structure which stores a snow particle's position                         */
/*---------------------------------------------------------------------------*/
typedef struct {
    unsigned int  offset;  // byte offset of particle (if this equals to
                           // FREE_ENTRY than this entry is free to re-use)
    unsigned char mask;    // bit mask of pixel
} particle;


/*---------------------------------------------------------------------------*/
/* global variables                                                          */
/*---------------------------------------------------------------------------*/
int            _ti92plus;          // needed to produce ti92plus code
int            _ti89;              // needed to produce ti89 code

unsigned char* plane     = 0;  // start address of bitplane
particle*      positions = 0;  // array of particle positions
int            count     = 0;  // number of actual snow particles


/*===========================================================================*/
/* fills particles array                                                     */
/*===========================================================================*/
void FillParticlesArray() {
    int           i;
    int           j=0;
    unsigned int  pix;
    unsigned int  offset;
    unsigned char mask;

    for (i=count;i<NR_PARTICLES;i++) {
        //----------------------------------------------
        // search for next free entry in particles array
        //----------------------------------------------
        while (j<NR_PARTICLES && positions[j].offset != FREE_ENTRY) j++;
        if (j>=NR_PARTICLES) return;
        
        //----------------------------------------------
        // select a random pixel and calculate its byte
        // offset and its bitmask
        //----------------------------------------------
        pix = random(RAND_RANGE);
        offset = pix / 8;
        mask   = 0x80 >> (pix % 8);

        //----------------------------------------------
        // if the randomly selected pixel is set, store
        // its values into the actual particles array
        // entry
        //----------------------------------------------
        if (*(plane+offset) & mask) {
            count++;
            positions[j].offset = offset;
            positions[j].mask   = mask;
        }
    }
}


/*===========================================================================*/
/* process particles in particles array                                      */
/*===========================================================================*/
void ProcessParticlesArray() {
    int i;

    for (i=0;i<NR_PARTICLES;i++) {
        //--------------------------------------------------
        // if this entry is not marked as free, check if the
        // pixel just below it is set. If there is no pixel
        // just below it, remove the actual pixel and move
        // it one line down
        //--------------------------------------------------
        if (positions[i].offset != FREE_ENTRY) {
            if ((*(plane + positions[i].offset + 30) & positions[i].mask) == 0) {
                *(plane + positions[i].offset) &= ~positions[i].mask;
                positions[i].offset += 30;
                *(plane + positions[i].offset) |= positions[i].mask;
                if (positions[i].offset / 30 >= 127) {
                    positions[i].offset = FREE_ENTRY;
                    count--;
                }
            }
            else {
                //-------------------------------------------
                // if there is a pixel exactly one line below
                // the actual one, mark the actual particles
                // array entry as free, because it cannot
                // "fall" down
                //-------------------------------------------
                positions[i].offset = FREE_ENTRY;
                count--;
            }
        }
    }
}


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    char* s1 = "Snow Effect Demo";
    char* s2 = "(c) TI-Chess Team 06/2000";
    int   i;

    //------------------------------------------------
    // try to allocate memory for particles
    //------------------------------------------------
    if ((positions = (particle*)malloc(sizeof(particle)*NR_PARTICLES))==0) {
        ClearScreen();
        DrawStr(0,0,"ERROR: too less mem",A_REPLACE);
        ngetchx();
        return;
    }

    //------------------------------------------------
    // mark all entries in the particles array as free
    // and initialize the global variables
    //------------------------------------------------
    for (i=0;i<NR_PARTICLES;i++) positions[i].offset = FREE_ENTRY;

    plane = LCD_MEM;
    count = 0;

    //---------------------------------------------
    // main loop
    //---------------------------------------------
    do {
        FontSetSys(F_8x10);
        DrawStr((ScrRect->xy.x1-DrawStrWidth(s1,F_8x10))/2,40,s1,A_REPLACE);
        FontSetSys(F_6x8);
        DrawStr((ScrRect->xy.x1-DrawStrWidth(s2,F_6x8))/2,60,s2,A_REPLACE);
        if (count < NR_PARTICLES-1) FillParticlesArray();
        ProcessParticlesArray();
    }
    while (!kbhit());
    ngetchx();

    //---------------------------------------------
    // cleanup and bye =8-)
    //---------------------------------------------
    free(positions);
}
