/******************************************************************************
*
* project name:    STARS
* initial date:    06/06/2000
* author:          thomas.nussbaumer@peem.com
* description:     starfield effect demonstration
*
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
* NOTE: depends on TIGCC Standard Library 2.0 to compile
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
*
* compile with: tigcc -O2 stars.c
*
* -----------------------------------------------------------------------------
* Revision History
* -----------------------------------------------------------------------------
*
* v1.00: 16/06/2000
*        initial version
*
*******************************************************************************/

#define SAVE_SCREEN      // restore screen after program end
#define USE_GRAY_PATCH   // use grayscale patch

#include <nostub.h>
#include <all.h>


/*---------------------------------------------------------------------------*/
/* some constants defining the look of the starfield                         */
/*---------------------------------------------------------------------------*/
#define NR_STARS     250
#define XOFFSET      (LCD_WIDTH/2)
#define YOFFSET      (LCD_HEIGHT/2)
#define ZOFFSET      200
#define MAX_STAR_X   16000
#define MAX_STAR_Y   16000
#define ZDECREMENT   4
#define BORDER_BLACK 100
#define BORDER_GRAY  140


/*---------------------------------------------------------------------------*/
/* structure which stores a star's location                                  */
/*---------------------------------------------------------------------------*/
typedef struct {
    int x,y,z;
    int offset;  // offset of old location for cleanup
    int mask;    // inverted mask of old location for cleanup
} star_type;


/*---------------------------------------------------------------------------*/
/* global variables                                                          */
/*---------------------------------------------------------------------------*/
int            _ti92plus;          // needed to produce ti92plus code
int            _ti89;              // needed to produce ti89 code

unsigned char* plane1        = 0;  // startadress of grayscale bitplane 1
unsigned char* plane2        = 0;  // startadress of grayscale bitplane 2
star_type*     stars         = 0;  // array of stars 

// array used to lookup mask values for single pixels of a byte in a bitplane
unsigned char mask_values[8] = { 0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x1 };


/*===========================================================================*/
/* initializes star with given index                                         */
/*===========================================================================*/
void CreateStar(int index) {
    stars[index].x=random(MAX_STAR_X)-MAX_STAR_X/2;
    stars[index].y=random(MAX_STAR_X)-MAX_STAR_Y/2;
    stars[index].z=ZOFFSET;
}


/*===========================================================================*/
/* delete star at old position/calculate new position/draw new position      */
/*===========================================================================*/
void DrawStars() {
    int           offset;
    unsigned char mask;
    int           drawx;
    int           drawy;
    int           i;
    
    for (i=0;i<NR_STARS;i++) {
        if (stars[i].offset != -1) {
            *(plane1+stars[i].offset) &= stars[i].mask;
            *(plane2+stars[i].offset) &= stars[i].mask;
            stars[i].offset = -1;
        }

        //---------------------------------
        // calculate new drawing x position 
        // and check if it out of screen
        //---------------------------------
        drawx  = stars[i].x / stars[i].z + XOFFSET;
        if (drawx < 0 || drawx >= LCD_WIDTH) {
            CreateStar(i);
            continue;
        }

        //---------------------------------
        // calculate new drawing y position 
        // and check if it out of screen
        //---------------------------------
        drawy  = stars[i].y / stars[i].z + YOFFSET;
        if (drawy < 0 || drawy >= LCD_HEIGHT) {
            CreateStar(i);
            continue;
        }
        
        //---------------------------------
        // calculate byte offset of pixel
        //---------------------------------
        offset = drawy * 30 + (drawx >> 3);
        
        //-----------------------------------
        // lookup mask value
        //-----------------------------------
        mask = mask_values[drawx % 8];

        //-----------------------------------
        // set pixel in plane 1 and/or plane2
        // depending on z value
        //-----------------------------------
        if (stars[i].z < BORDER_BLACK) {
            *(plane1+offset) |= mask;
            *(plane2+offset) |= mask;
        }

        else if (stars[i].z > BORDER_GRAY) {        
            *(plane1+offset) |= mask;
        }

        else {
            *(plane2+offset) |= mask;
        }

        //-----------------------------------
        // store offset and mask so we can 
        // easily remove this pixel later
        //-----------------------------------
        stars[i].mask   = ~mask;
        stars[i].offset = offset;

        //-----------------------------------
        // decrement z and check if we must
        // generate a new star
        //-----------------------------------
        stars[i].z -= ZDECREMENT;
        if (stars[i].z <= 0) CreateStar(i);
    }    
}


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    char* s1 = "Starfield Demo";
    char* s2 = "(c) TI-Chess Team 06/2000";
    int   i;
    
    //---------------------------------------------
    // try to allocate memory for stars
    //---------------------------------------------
    if ((stars = (star_type*)malloc(sizeof(star_type)*NR_STARS))==0) {
        ClearScreen();
        DrawStr(0,0,"ERROR: too less mem",A_REPLACE);
        ngetchx();
        return;
    }

    //---------------------------------------------
    // setup initial star positions
    //---------------------------------------------
    for (i=0;i<NR_STARS;i++) {
        stars[i].x=random(MAX_STAR_X)-MAX_STAR_X/2;
        stars[i].y=random(MAX_STAR_X)-MAX_STAR_X/2;
        stars[i].z=random(ZOFFSET)+1;
        stars[i].offset = -1;
    }
    
    //---------------------------------------------
    // turn on grayscale graphics subsystem
    //---------------------------------------------
    if (!GrayMode(GRAY_ON)) {
        free(stars);
        return;
    }

    //---------------------------------------------
    // clear both planes 
    //---------------------------------------------
    SetPlane(0);
    ClearScreen();
    SetPlane(1);
    ClearScreen();

    //---------------------------------------------
    // fetch start addresses of planes
    //---------------------------------------------
    plane1 = GetPlane(0);
    plane2 = GetPlane(1);

    //---------------------------------------------
    // main loop
    //---------------------------------------------
    do {
        DrawStars();
        //-------------
        // output title
        //-------------
        for (i=0;i<2;i++) {
            SetPlane(i);
            FontSetSys(F_8x10);
            DrawStr((ScrRect->xy.x1-DrawStrWidth(s1,F_8x10))/2,10,s1,A_NORMAL);
            FontSetSys(F_4x6);
            DrawStr((ScrRect->xy.x1-DrawStrWidth(s2,F_4x6))/2,30,s2,A_NORMAL);
        }
    } 
    while (!kbhit());
    ngetchx();

    //---------------------------------------------
    // cleanup and bye =8-)
    //---------------------------------------------
    GrayMode(GRAY_OFF);
    free(stars);
}

